@@ -1,13 +1,14 @@
 """
 Definitions of various types of fields. Supports JSON draft4 types.
 """
+import enum
 import re
 from abc import ABC
 from collections import OrderedDict
 from functools import reduce
-from decimal import Decimal, getcontext, InvalidOperation
+from decimal import Decimal, InvalidOperation
 
-from typedpy.structures import Field, Structure, TypedField, ClassReference, StructMeta
+from typedpy.structures import Field, Structure, TypedField, ClassReference, StructMeta, is_function_returning_field
 
 
 def _map_to_field(item):
@@ -21,6 +22,9 @@ def _map_to_field(item):
         raise TypeError("Expected a Field/Structure class or Field instance")
 
 
+def wrap_val(v): return "'{}'".format(v) if isinstance(v, str) else v
+
+
 class StructureReference(Field):
     """
     A Field that is an embedded structure within other structure. Allows to create hierarchy.
@@ -98,7 +102,7 @@ class Number(Field):
             return isinstance(val, (float, int, Decimal))
 
         def err_prefix():
-            return "{}: ".format(self._name) if self._name else ""
+            return "{}: Got {}; ".format(self._name, wrap_val(value)) if self._name else ""
 
         if not is_number(value):
             raise TypeError("{}Expected a number".format(err_prefix()))
@@ -112,11 +116,11 @@ class Number(Field):
                 err_prefix(), self.minimum))
         if is_number(self.maximum):
             if self.exclusiveMaximum and self.maximum == value:
-                raise ValueError("{}Expected a maxmimum of less than {}".format(
+                raise ValueError("{}Expected a maximum of less than {}".format(
                     err_prefix(), self.maximum))
             else:
                 if self.maximum < value:
-                    raise ValueError("{}Expected a maxmimum of {}".format(
+                    raise ValueError("{}Expected a maximum of {}".format(
                         err_prefix(), self.maximum))
 
     def _validate(self, value):
@@ -154,6 +158,10 @@ class DecimalNumber(Number):
         super().__set__(instance, value)
 
 
+class StructureClass(TypedField):
+    _ty = StructMeta
+
+
 class String(TypedField):
     """
       A string value. Accepts input of `str`
@@ -184,12 +192,12 @@ class String(TypedField):
     @staticmethod
     def _validate_static(self, value):
         def err_prefix():
-            return "{}: ".format(self._name) if self._name else ""
+            return "{}: Got {}; ".format(self._name, wrap_val(value)) if self._name else ""
 
         if not isinstance(value, str):
             raise TypeError("{}Expected a string".format(err_prefix()))
         if self.maxLength is not None and len(value) > self.maxLength:
-            raise ValueError("{}Expected a maxmimum length of {}".format(
+            raise ValueError("{}Expected a maximum length of {}".format(
                 err_prefix(), self.maxLength))
         if self.minLength is not None and len(value) < self.minLength:
             raise ValueError("{}Expected a minimum length of {}".format(
@@ -203,6 +211,25 @@ class String(TypedField):
         super().__set__(instance, value)
 
 
+def _foo(): pass
+
+
+class Function(Field):
+    """
+       A function. Note that this can't be any callable (it can't be a class, for example), but a real function
+    """
+    def __set__(self, instance, value):
+        def is_function(f):
+            return type(f) == type(_foo) or type(f) == type(open)
+
+        def err_prefix():
+            return "{}: Got {}; ".format(self._name, wrap_val(value)) if self._name else ""
+
+        if not is_function(value):
+            raise TypeError("{}Expected a function".format(err_prefix()))
+        super().__set__(instance, value)
+
+
 class Anything(Field):
     """
     A field that can contain anything (similar to "any" in Typescript).
@@ -349,10 +376,12 @@ class _CollectionMeta(type):
         def validate_and_get_field(val):
             if isinstance(val, Field):
                 return val
-            elif Field in val.__mro__:
+            elif Field in getattr(val, '__mro__', {}):
                 return val()
-            elif Structure in val.__mro__:
+            elif Structure in getattr(val, '__mro__', {}):
                 return ClassReference(val)
+            elif is_function_returning_field(val):
+                return val()
             else:
                 raise TypeError("Expected a Field class or instance")
 
@@ -364,6 +393,8 @@ class _CollectionMeta(type):
 
 class _EnumMeta(type):
     def __getitem__(cls, values):
+        if isinstance(values, (type,)) and issubclass(values, (enum.Enum,)):
+            return cls(values=values)
         return cls(values=list(values))
 
 
@@ -372,10 +403,12 @@ class _JSONSchemaDraft4ReuseMeta(type):
         def validate_and_get_field(val):
             if isinstance(val, Field):
                 return val
-            elif Field in val.__mro__:
+            elif Field in getattr(val, '__mro__', {}):
                 return val()
-            elif Structure in val.__mro__:
+            elif Structure in getattr(val, '__mro__', {}):
                 return ClassReference(val)
+            elif is_function_returning_field(val):
+                return val()
             else:
                 raise TypeError("Expected a Field class or instance")
 
@@ -439,7 +472,7 @@ class Set(SizedCollection, TypedField, metaclass=_CollectionMeta):
 
     def __set__(self, instance, value):
         if not isinstance(value, set):
-            raise TypeError("%s: Expected %s" % (self._name, set))
+            raise TypeError("{}: Got {}; Expected {}".format(self._name, wrap_val(value), set))
         self.validate_size(value, self._name)
         if self.items is not None:
             temp_st = Structure()
@@ -492,6 +525,7 @@ class Map(SizedCollection, TypedField, metaclass=_CollectionMeta):
             if isinstance(key_field, TypedField) and not getattr(getattr(key_field, '_ty'), '__hash__'):
                 raise TypeError("Key field of type {}, with underlying type of {} is not hashable".format(
                     key_field, getattr(key_field, '_ty')))
+        self._custom_deep_copy_implementation = True
         super().__init__(*args, **kwargs)
 
     def __set__(self, instance, value):
@@ -512,7 +546,6 @@ class Map(SizedCollection, TypedField, metaclass=_CollectionMeta):
                 res[getattr(temp_st, getattr(key_field, '_name'))] = getattr(
                     temp_st, getattr(value_field, '_name'))
             value = res
-
         super().__set__(instance, _DictStruct(self, instance, value))
 
 
@@ -590,13 +623,17 @@ class Array(SizedCollection, TypedField, metaclass=_CollectionMeta):
             elif isinstance(self.items, list):
                 additional_properties_forbidden = self.additionalItems is not None and \
                                                   self.additionalItems is False
-                if len(self.items) > len(value) or \
-                        (additional_properties_forbidden and len(self.items) > len(value)):
-                    raise ValueError("{}: Expected an array of length {}".format(
-                        self._name, len(self.items)))
+                if not getattr(instance, '_skip_validation', False):
+                    if len(self.items) > len(value) or \
+                            (additional_properties_forbidden and len(self.items) > len(value)):
+                        raise ValueError("{}: Got {}; Expected an array of length {}".format(
+                            self._name, value, len(self.items)))
                 temp_st = Structure()
+                temp_st._skip_validation = getattr(instance, '_skip_validation', False)
                 res = []
                 for ind, item in enumerate(self.items):
+                    if ind >= len(value):
+                        continue
                     setattr(item, '_name', self._name + "_{}".format(str(ind)))
                     item.__set__(temp_st, value[ind])
                     res.append(getattr(temp_st, getattr(item, '_name')))
@@ -608,13 +645,13 @@ class Array(SizedCollection, TypedField, metaclass=_CollectionMeta):
 
 def verify_type_and_uniqueness(the_type, value, name, has_unique_items):
     if not isinstance(value, the_type):
-        raise TypeError("%s: Expected %s" % (name, the_type))
+        raise TypeError("{}: Got {}; Expected {}".format(name, wrap_val(value), str(the_type)))
     if has_unique_items:
         unique = reduce(lambda unique_vals, x: unique_vals.append(x) or
                         unique_vals if x not in unique_vals
                         else unique_vals, value, [])
         if len(unique) < len(value):
-            raise ValueError("{}: Expected unique items".format(name))
+            raise ValueError("{}: Got {}; Expected unique items".format(name, wrap_val(value)))
 
 
 class Tuple(TypedField, metaclass=_CollectionMeta):
@@ -686,8 +723,8 @@ class Tuple(TypedField, metaclass=_CollectionMeta):
     def __set__(self, instance, value):
         verify_type_and_uniqueness(tuple, value, self._name, self.uniqueItems)
         if len(self.items) != len(value) and len(self.items) > 1:
-            raise ValueError("{}: Expected a tuple of length {}".format(
-                self._name, len(self.items)))
+            raise ValueError("{}: Got {}; Expected a tuple of length {}".format(
+                self._name, wrap_val(value), len(self.items)))
 
         temp_st = Structure()
         res = []
@@ -710,18 +747,44 @@ class Enum(Field, metaclass=_EnumMeta):
              values(`list` or `set` or `tuple`):
                  allowed values. Can be of any type
 
+        Examples:
+
+        .. code-block:: python
+
+           class Values(enum.Enum):
+                ABC = enum.auto()
+                DEF = enum.auto()
+                GHI = enum.auto()
+
+           class Example(Structure):
+              arr = Array[Enum[Values]]
+              e = Enum['abc', 'x', 'def', 3]
+
+           Example(arr=[Values.ABC, 'DEF'],e=3)
+
     """
 
     def __init__(self, *args, values, **kwargs):
+        self._is_enum = isinstance(values, (type,)) and issubclass(values, enum.Enum)
+        if self._is_enum:
+            self._enum_class = values
         self.values = values
         super().__init__(*args, **kwargs)
 
     def _validate(self, value):
-        if value not in self.values:
+        if self._is_enum:
+            enum_names = {v.name for v in self._enum_class}
+            if value not in enum_names and not isinstance(value, (self._enum_class,)):
+                raise ValueError('{}: Must be a value of {}'.format(self._name, self._enum_class))
+
+        elif value not in self.values:
             raise ValueError('{}: Must be one of {}'.format(self._name, self.values))
 
     def __set__(self, instance, value):
         self._validate(value)
+        if self._is_enum:
+            if isinstance(value, (str,)):
+                value = self._enum_class[value]
         super().__set__(instance, value)
 
 
@@ -857,7 +920,7 @@ class AnyOf(MultiFieldWrapper, Field, metaclass=_JSONSchemaDraft4ReuseMeta):
             except ValueError:
                 pass
         if not matched:
-            raise ValueError("{}: {} Did not match any field option".format(self._name, value))
+            raise ValueError("{}: {} Did not match any field option".format(self._name, wrap_val(value)))
         super().__set__(instance, value)
 
     def __str__(self):
@@ -895,9 +958,9 @@ class OneOf(MultiFieldWrapper, Field, metaclass=_JSONSchemaDraft4ReuseMeta):
             except ValueError:
                 pass
         if not matched:
-            raise ValueError("{}: Did not match any field option".format(self._name))
+            raise ValueError("{}: Got {}; Did not match any field option".format(self._name, value))
         if matched > 1:
-            raise ValueError("{}: Matched more than one field option".format(self._name))
+            raise ValueError("{}: Got {}; Matched more than one field option".format(self._name, value))
         super().__set__(instance, value)
 
     def __str__(self):
@@ -934,8 +997,8 @@ class NotField(MultiFieldWrapper, Field, metaclass=_JSONSchemaDraft4ReuseMeta):
             except ValueError:
                 pass
             else:
-                raise ValueError("{}: Expected not to match any field definition".
-                                 format(self._name))
+                raise ValueError("{}: Got {}; Expected not to match any field definition".
+                                 format(self._name, wrap_val(value)))
         super().__set__(instance, value)
 
     def __str__(self):
@@ -997,3 +1060,13 @@ class SerializableField(ABC):
 
     def deserialize(self, value): return value
 
+
+class ExceptionField(TypedField, SerializableField):
+    """
+    As Exception. This is serialized as the string representation of the exception.
+    It does not support deserialization.
+    """
+    _ty = Exception
+
+    def serialize(self, value):
+        return str(value)
