@@ -2,6 +2,7 @@
 The Skeleton classes to support strictly defined structures:
 Structure, Field, StructureReference, ClassReference, TypedField
 """
+from copy import deepcopy
 from collections import OrderedDict
 from inspect import Signature, Parameter
 import sys
@@ -71,7 +72,7 @@ def get_base_info(bases):
     return bases_params, bases_required
 
 
-class Field(object):
+class Field:
     """
     Base class for a field(i.e. property) in a structure.
     Should not be used directly by developers.
@@ -86,14 +87,16 @@ class Field(object):
     def __get__(self, instance, owner):
         if instance and self._name not in instance.__dict__:
             return self._default
-        else:
-            return instance.__dict__[self._name] if instance else owner.__dict__[self._name]
+        return instance.__dict__[self._name] if instance else owner.__dict__[self._name]
 
     def __set__(self, instance, value):
-        if getattr(self, '_immutable', False) \
-                and self._name in instance.__dict__:
+        if getattr(self, '_immutable', False) and self._name in instance.__dict__:
             raise ValueError("{}: Field is immutable".format(self._name))
-        instance.__dict__[self._name] = value
+        if getattr(self, '_immutable', False) and \
+                not getattr(self, '_custom_deep_copy_implementation', False):
+            instance.__dict__[self._name] = deepcopy(value)
+        else:
+            instance.__dict__[self._name] = value
         if getattr(instance, '_instantiated', False):
             instance.__validate__()
 
@@ -119,6 +122,19 @@ class Field(object):
         return '<{}{}>'.format(name, propst)
 
 
+# noinspection PyBroadException
+def is_function_returning_field(field_definition_candidate):
+    python_ver_higher_than_36 = sys.version_info[0:2] != (3, 6)
+    if callable(field_definition_candidate) and python_ver_higher_than_36:
+        try:
+            return_value = get_type_hints(field_definition_candidate).get("return", None)
+            return return_value == Field or \
+                   Field in getattr(return_value.__args__[0], '__mro__', [])
+        except Exception:
+            return False
+    return False
+
+
 class StructMeta(type):
     """
     Metaclass for Structure. Manipulates it to ensure the fields are set up correctly.
@@ -129,19 +145,9 @@ class StructMeta(type):
         return OrderedDict()
 
     def __new__(mcs, name, bases, cls_dict):
-        # noinspection PyBroadException
-        def is_function_returning_field(field_definition_candidate):
-            if callable(field_definition_candidate) and sys.version_info[0:2] != (3,6):
-                try:
-                    return_value = get_type_hints(field_definition_candidate).get("return", None)
-                    return Field in getattr(return_value.__args__[0], '__mro__', [])
-                except Exception:
-                    return False
-            return False
-
         bases_params, bases_required = get_base_info(bases)
         for key, val in cls_dict.items():
-            if not key.startswith('_') and not isinstance(val, Field) and \
+            if key not in {'_required', '_additionalProperties', '_immutable'} and not isinstance(val, Field) and \
                     (Field in getattr(val, '__mro__', []) or is_function_returning_field(val)):
                 cls_dict[key] = val()
         for key, val in cls_dict.items():
@@ -149,6 +155,8 @@ class StructMeta(type):
                 cls_dict[key] = ClassReference(val)
         fields = [key for key, val in cls_dict.items() if isinstance(val, Field)]
         for field_name in fields:
+            if field_name.startswith('_') or field_name == 'kwargs':
+                raise ValueError("{}: invalid field name".format(field_name))
             setattr(cls_dict[field_name], '_name', field_name)
         clsobj = super().__new__(mcs, name, bases, dict(cls_dict))
         clsobj._fields = fields
@@ -227,8 +235,10 @@ class Structure(metaclass=StructMeta):
         self._instantiated = True
 
     def __setattr__(self, key, value):
-        if getattr(self, '_immutable', False) and key in self.__dict__:
-            raise ValueError("Structure is immutable")
+        if getattr(self, '_immutable', False):
+            if key in self.__dict__:
+                raise ValueError("Structure is immutable")
+            value = deepcopy(value)
         super().__setattr__(key, value)
 
     def __str__(self):
@@ -241,13 +251,13 @@ class Structure(metaclass=StructMeta):
             return ','.join(as_strings)
 
         def to_str(val):
-            if isinstance(val, (list)):
+            if isinstance(val, list):
                 return '[{}]'.format(list_to_str(val))
-            elif isinstance(val, tuple):
+            if isinstance(val, tuple):
                 return '({})'.format(list_to_str(val))
-            elif isinstance(val, set):
+            if isinstance(val, set):
                 return '{{{}}}'.format(list_to_str(val))
-            elif isinstance(val, dict):
+            if isinstance(val, dict):
                 return '{{{}}}'.format(dict_to_str(val))
             return str(val)
 
